import shutil
from pathlib import Path

import pytest

from bumpversion.scm.hg import Mercurial
from bumpversion.scm.models import SCMConfig
from tests.conftest import inside_dir


def test_hg_is_not_available(tmp_path: Path, scm_config: SCMConfig) -> None:
    """Should return false if it is not a mercurial repo."""
    with inside_dir(tmp_path):
        assert not Mercurial(scm_config).is_available()


@pytest.mark.skipif(not shutil.which("hg"), reason="Mercurial is not available.")
def test_hg_is_available(hg_repo: Path, scm_config: SCMConfig) -> None:
    """Should return false if it is not a mercurial repo."""
    with inside_dir(hg_repo):
        assert Mercurial(scm_config).is_available()



"""Data models for source code management functions."""

import re
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any, MutableMapping, Optional, Protocol

from bumpversion.config import Config
from bumpversion.ui import get_indented_logger
from bumpversion.utils import extract_regex_flags

logger = get_indented_logger(__name__)


@dataclass
class SCMConfig:
    """Configuration for source code management functions."""

    tag: bool
    sign_tags: bool
    tag_name: str
    allow_dirty: bool
    commit: bool
    message: str
    parse_pattern: str
    tag_message: Optional[str] = None
    commit_args: Optional[str] = None
    moveable_tags: list[str] = field(default_factory=list)

    def get_version_from_tag(self, tag: str) -> Optional[str]:
        """Return the version from a tag."""
        version_pattern = self.parse_pattern.replace("\\\\", "\\")
        version_pattern, regex_flags = extract_regex_flags(version_pattern)
        parts = self.tag_name.split("{new_version}", maxsplit=1)
        prefix = parts[0]
        # suffix = parts[1]
        rep = f"{regex_flags}{re.escape(prefix)}(?P<current_version>{version_pattern})"
        tag_regex = re.compile(rep)
        return match["current_version"] if (match := tag_regex.search(tag)) else None

    @classmethod
    def from_config(cls, config: Config) -> "SCMConfig":
        """Return a SCMConfig from a Config object."""
        return cls(
            tag=config.tag,
            tag_name=config.tag_name,
            sign_tags=config.sign_tags,
            allow_dirty=config.allow_dirty,
            commit=config.commit,
            message=config.message,
            parse_pattern=config.parse,
            tag_message=config.tag_message,
            commit_args=config.commit_args,
            moveable_tags=config.moveable_tags,
        )


@dataclass
class LatestTagInfo:
    """Information about the latest tag."""

    commit_sha: Optional[str] = None
    distance_to_latest_tag: int = 0
    current_version: Optional[str] = None
    current_tag: Optional[str] = None
    branch_name: Optional[str] = None
    short_branch_name: Optional[str] = None
    repository_root: Optional[Path] = None
    dirty: Optional[bool] = None


class SCMTool(Protocol):
    """Protocol for source code management tools."""

    def __init__(self, config: SCMConfig): ...

    def is_available(self) -> bool:
        """Return whether the SCM tool is available."""
        ...

    def latest_tag_info(self) -> LatestTagInfo:
        """Return the latest tag information."""
        ...

    def add_path(self, path: str | Path) -> None:
        """Add a path to the pending commit."""
        ...

    def get_all_tags(self) -> list[str]:
        """Return all tags in the SCM."""
        ...

    def commit_and_tag(self, files: list[Path | str], context: MutableMapping, dry_run: bool = False) -> None:
        """Commit and tag files to the repository using the configuration."""
        ...

    def assert_nondirty(self) -> None:
        """
        Asserts that the repository is not dirty.

        Raises:
            DirtyWorkingDirectoryError: If the repository is not clean.
        """
        ...


class DefaultSCMTool:
    """Default implementation of the SCMTool protocol."""

    def __init__(self, config: SCMConfig):
        self.config = config

    def is_available(self) -> bool:
        """Return whether the SCM tool is available."""
        return True

    def latest_tag_info(self) -> LatestTagInfo:
        """Return the latest tag information."""
        return LatestTagInfo()

    def add_path(self, path: str | Path) -> None:
        """Add a path to the pending commit."""
        logger.debug("No source code management system configured. Skipping adding path '%s'.", path)

    def get_all_tags(self) -> list[str]:
        """Return all tags in the SCM."""
        return []

    def commit_and_tag(self, files: list[Path | str], context: MutableMapping, dry_run: bool = False) -> None:
        """Pretend to commit and tag files to the repository using the configuration."""
        logger.debug("No source code management system configured. Skipping committing and tagging.")

    def assert_nondirty(self) -> None:
        """Always says the repository is clean."""
        return


class SCMInfo:
    """Information about the current source code manager and state."""

    def __init__(self, config: SCMConfig):
        self.config = config
        self.commit_sha: Optional[str] = None
        self.distance_to_latest_tag: int = 0
        self.current_version: Optional[str] = None
        self.current_tag: Optional[str] = None
        self.branch_name: Optional[str] = None
        self.short_branch_name: Optional[str] = None
        self.repository_root: Optional[Path] = None
        self.dirty: Optional[bool] = None
        self.tool: SCMTool = DefaultSCMTool(config)
        self._set_from_scm_tool()

    def as_dict(self) -> dict[str, Any]:
        """Return the information as a dict."""
        return {
            "commit_sha": self.commit_sha,
            "distance_to_latest_tag": self.distance_to_latest_tag,
            "current_version": self.current_version,
            "current_tag": self.current_tag,
            "branch_name": self.branch_name,
            "short_branch_name": self.short_branch_name,
            "repository_root": self.repository_root,
            "dirty": self.dirty,
            "tool": self.tool,
        }

    def _set_from_scm_tool(self):
        """Set information from the source code management tool."""
        from bumpversion.scm.git import Git
        from bumpversion.scm.hg import Mercurial

        git = Git(self.config)
        hg = Mercurial(self.config)
        if git.is_available():
            self.tool = git
        elif hg.is_available():
            self.tool = hg
        self._update_from_latest_tag_info(self.tool.latest_tag_info())

    def _update_from_latest_tag_info(self, latest_tag_info: LatestTagInfo):
        """Update information from the latest tag information."""
        for attr_name, value in asdict(latest_tag_info).items():
            setattr(self, attr_name, value)

    def path_in_repo(self, path: Path | str) -> bool:
        """Return whether a path is inside this repository."""
        if self.repository_root is None:
            return True
        elif not Path(path).is_absolute():
            return True
        return str(path).startswith(str(self.repository_root))

    def commit_and_tag(
        self,
        files: list[str | Path],
        context: MutableMapping,
        dry_run: bool = False,
    ) -> None:
        """Commit the files to the source code management system."""
        logger.indent()

        if self.config.commit:
            self._commit(files, context, dry_run)
        else:
            logger.info("Would not commit")

        if self.config.tag:
            self._tag(context, dry_run)
        else:
            logger.info("Would not tag")

        self.tool.commit_and_tag(files=files, context=context, dry_run=dry_run)
        logger.dedent()

    def _commit(self, files: list[Path | str], context: MutableMapping, dry_run: bool = False) -> None:
        """Commit the files to the source code management system."""
        do_commit = not dry_run
        logger.info("%s the commit", "Preparing" if do_commit else "Would prepare")

        for path in files:
            logger.info("%s changes in file '%s' to repository", "Adding" if do_commit else "Would add", path)

        commit_message = self.config.message.format(**context)
        logger.info(
            "%s to repository with message '%s'", "Committing" if do_commit else "Would commit", commit_message
        )

    def _tag(self, context: MutableMapping, dry_run: bool = False) -> None:
        """Tag the current commit in the source code management system."""
        sign_tags = self.config.sign_tags
        tag_name = self.config.tag_name.format(**context)
        tag_message = self.config.tag_message.format(**context)
        existing_tags = self.tool.get_all_tags()
        do_tag = not dry_run

        if tag_name in existing_tags:
            logger.warning("Tag '%s' already exists. Will not tag.", tag_name)
            return

        logger.info(
            "%s '%s' %s to repository and %s",
            "Adding tag" if do_tag else "Would add tag",
            tag_name,
            f"with message '{tag_message}'" if tag_message else "without message",
            "signing" if sign_tags else "not signing",
        )

        for moveable_tag in self.config.moveable_tags:
            tag_name = moveable_tag.format(**context)
            logger.info("%s moveable tag '%s' in repository", "Tagging" if do_tag else "Would tag", tag_name)



"""File processing tests."""

import shutil
from datetime import datetime
from difflib import context_diff
from pathlib import Path
from textwrap import dedent

import pytest
import tomlkit
from pytest import param

from bumpversion import bump, config, exceptions, files
from bumpversion.config.models import FileChange
from bumpversion.context import get_context
from bumpversion.exceptions import VersionNotFoundError
from bumpversion.versioning.version_config import VersionConfig
from tests.conftest import get_config_data, inside_dir


def test_modify_files_raises_correct_missing_version_string(tmp_path: Path, fixtures_path: Path):
    """When a file is missing the version string, the error should indicate the correct serialization missing."""
    csharp_path = fixtures_path / "csharp"
    dst_path: Path = shutil.copytree(csharp_path, tmp_path / "csharp")

    with inside_dir(dst_path):
        conf = config.get_configuration(config_file=dst_path.joinpath(".bumpversion.toml"))
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)
        dst_path.joinpath("Version.csv").write_text("1;3-1;0;rc;build.1031", encoding="utf-8")

        major_version = current_version.bump("patch")
        ctx = get_context(conf)

        configured_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        with pytest.raises(VersionNotFoundError) as e:
            files.modify_files(configured_files, current_version, major_version, ctx)
            assert e.message.startswith("Did not find '1;3;1;0;rc;build.1031'")


def test_do_bump_uses_correct_serialization_for_tag_and_commit_messages(git_repo: Path, fixtures_path: Path, caplog):
    """The tag and commit messages should use the root configured serialization."""
    import subprocess
    import logging

    caplog.set_level(logging.INFO)

    csharp_path = fixtures_path / "csharp"
    dst_path: Path = shutil.copytree(csharp_path, git_repo / "csharp")
    subprocess.check_call(["git", "add", "."], cwd=git_repo)
    subprocess.check_call(["git", "commit", "-m", "Initial commit"], cwd=git_repo)
    subprocess.check_call(["git", "tag", "3.1.0-rc+build.1031"], cwd=git_repo)

    with inside_dir(dst_path):
        config_file = dst_path.joinpath(".bumpversion.toml")
        conf = config.get_configuration(config_file=config_file)
        bump.do_bump("patch", None, conf, config_file, dry_run=True)

    assert "Bump version: 3.1.0-rc+build.1031 -> 3.1.1" in caplog.text


def test_modify_files_non_matching_search_does_not_modify_file(tmp_path: Path):
    changelog_path = tmp_path / "CHANGELOG.md"
    changelog_content = dedent(
        """
    # Unreleased

    * bullet point A

    # Release v'older' (2019-09-17)

    * bullet point B
    """
    )
    changelog_path.write_text(changelog_content, encoding="utf-8")

    overrides = {
        "current_version": "1.0.3",
        "files": [
            {
                "filename": str(changelog_path),
                "search": "Not-yet-released",
                "replace": "Release v{new_version} ({now:%Y-%m-%d})",
            }
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    configured_files = files.resolve_file_config(conf.files, version_config)
    with pytest.raises(exceptions.VersionNotFoundError, match="Did not find 'Not-yet-released' in file:"):
        files.modify_files(configured_files, current_version, new_version, get_context(conf))

    assert changelog_path.read_text() == changelog_content


@pytest.mark.parametrize(
    ["global_value", "file_value", "should_raise"],
    [
        param(True, True, False, id="ignore global and file"),
        param(True, False, True, id="ignore global only"),
        param(False, True, False, id="ignore file only"),
        param(False, False, True, id="ignore none"),
    ],
)
def test_modify_files_ignore_missing_version(
    global_value: bool, file_value: bool, should_raise: bool, tmp_path: Path
) -> None:
    """If the version is not found in the file, do nothing."""
    # Arrange
    version_path = tmp_path / Path("VERSION")
    version_path.write_text("1.2.3", encoding="utf-8")

    overrides = {
        "current_version": "1.2.5",
        "ignore_missing_version": global_value,
        "files": [{"filename": str(version_path), "ignore_missing_version": file_value}],
    }
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        if should_raise:
            with pytest.raises(VersionNotFoundError):
                files.modify_files(cfg_files, current_version, new_version, get_context(conf))
        else:
            files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    assert version_path.read_text() == "1.2.3"


def test_modify_files_regex_search(tmp_path: Path) -> None:
    """A regex search string is found and replaced."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("Release: 1234-56-78 '1.2.3'", encoding="utf-8")

    overrides = {
        "regex": True,
        "current_version": "1.2.3",
        "search": r"Release: \d{{4}}-\d{{2}}-\d{{2}} '{current_version}'",
        "replace": r"Release {now:%Y-%m-%d} '{new_version}'",
        "files": [{"filename": str(version_path)}],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

    # Act
    files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    now = datetime.now().isoformat()[:10]
    assert version_path.read_text() == f"Release {now} '1.2.4'"


def test_modify_files_regex_search_with_escaped_chars(tmp_path: Path) -> None:
    """A search that uses special characters is not treated as a regex."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("## [Release] 1.2.3 1234-56-78", encoding="utf-8")

    overrides = {
        "regex": True,
        "current_version": "1.2.3",
        "search": r"## \[Release\] {current_version} \d{{4}}-\d{{2}}-\d{{2}}",
        "replace": r"## [Release] {new_version} {now:%Y-%m-%d}",
        "files": [{"filename": str(version_path)}],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

    # Act
    files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    now = datetime.now().isoformat()[:10]
    assert version_path.read_text() == f"## [Release] 1.2.4 {now}"


def test_modify_files_regex_search_in_toml(tmp_path: Path, fixtures_path: Path) -> None:
    """Tests how-to doc 'update-a-date.md'."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("__date__ = '1234-56-78'\n__version__ = '1.2.3'", encoding="utf-8")
    config_path = tmp_path / ".bumpversion.toml"
    shutil.copyfile(fixtures_path / "replace-date-config.toml", config_path)
    with inside_dir(tmp_path):
        conf = config.get_configuration(config_file=config_path)
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)

        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    now = datetime.now().isoformat()[:10]
    assert version_path.read_text() == f"__date__ = '{now}'\n__version__ = '1.2.4'"


def test_modify_files_regex_search_with_caret(tmp_path: Path, fixtures_path: Path) -> None:
    """A search that uses a caret to indicate the beginning of the line works correctly."""
    # Arrange
    config_path = tmp_path / ".bumpversion.toml"
    thingy_path = tmp_path / "thingy.yaml"
    shutil.copyfile(fixtures_path / "regex_with_caret.yaml", thingy_path)
    shutil.copyfile(fixtures_path / "regex_with_caret_config.toml", config_path)

    conf = config.get_configuration(config_file=config_path)
    version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
    current_version = version_config.parse(conf.current_version)
    new_version = current_version.bump("patch")

    with inside_dir(tmp_path):
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    assert (
        thingy_path.read_text() == "version: 1.0.1\ndependencies:\n- name: kube-prometheus-stack\n  version: 1.0.0\n"
    )


def test_modify_files_bad_regex_search(tmp_path: Path, caplog) -> None:
    """A search string not meant to be a regex is still found and replaced."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("Score: A+ ( '1.2.3'", encoding="utf-8")

    overrides = {
        "regex": True,
        "current_version": "1.2.3",
        "search": r"Score: A+ ( '{current_version}'",
        "replace": r"Score: A+ ( '{new_version}'",
        "files": [{"filename": str(version_path)}],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

    # Act
    files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    assert version_path.read_text() == "Score: A+ ( '1.2.4'"
    assert "Invalid regex" in caplog.text


class TestDataFileUpdater:
    """Tests for the DataFileUpdater class."""

    def test_update_file_does_not_modify_non_toml_files(self, tmp_path: Path) -> None:
        """A non-TOML file is not modified."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("1.2.3", encoding="utf-8")

        overrides = {"current_version": "1.2.3", "files": [{"filename": str(version_path)}]}
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        datafile_config = FileChange(
            filename=str(version_path),
            key_path="",
            search=conf.search,
            replace=conf.replace,
            regex=conf.regex,
            ignore_missing_version=conf.ignore_missing_version,
            ignore_missing_file=conf.ignore_missing_files,
            serialize=conf.serialize,
            parse=conf.parse,
        )

        # Act
        files.DataFileUpdater(datafile_config, version_config.part_configs).update_file(
            current_version, new_version, get_context(conf)
        )

        # Assert
        assert version_path.read_text() == "1.2.3"

    def test_update_replaces_key(self, tmp_path: Path, fixtures_path: Path) -> None:
        """A key specific key is replaced and nothing else is touched."""
        # Arrange
        config_path = tmp_path / "pyproject.toml"
        fixture_path = fixtures_path / "partial_version_strings.toml"
        shutil.copy(fixture_path, config_path)

        contents_before = config_path.read_text()
        conf = config.get_configuration(config_file=config_path, files=[{"filename": str(config_path)}])
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)
        new_version = current_version.bump("minor")
        datafile_config = FileChange(
            filename=str(config_path),
            key_path="tool.bumpversion.current_version",
            search=conf.search,
            replace=conf.replace,
            regex=conf.regex,
            ignore_missing_version=conf.ignore_missing_version,
            ignore_missing_file=conf.ignore_missing_files,
            serialize=conf.serialize,
            parse=conf.parse,
        )

        # Act
        files.DataFileUpdater(datafile_config, version_config.part_configs).update_file(
            current_version, new_version, get_context(conf)
        )

        # Assert
        contents_after = config_path.read_text()
        toml_data = tomlkit.parse(config_path.read_text()).unwrap()
        actual_difference = list(
            context_diff(
                contents_before.splitlines(),
                contents_after.splitlines(),
                fromfile="before",
                tofile="after",
                n=0,
                lineterm="",
            )
        )
        expected_difference = [
            "*** before",
            "--- after",
            "***************",
            "*** 28 ****",
            '! current_version = "0.0.2"',
            "--- 28 ----",
            '! current_version = "0.1.0"',
        ]
        assert actual_difference == expected_difference
        assert toml_data["tool"]["pdm"]["dev-dependencies"]["lint"] == ["ruff==0.0.292"]
        assert toml_data["tool"]["bumpversion"]["current_version"] == "0.1.0"



"""Tests for the config.__init__ module."""

from unittest.mock import Mock

from bumpversion.config import Config, check_current_version
from bumpversion.scm.models import SCMConfig, SCMInfo


class TestCheckCurrentVersion:
    """
    Tests for the check_current_version function.

    - tag may not match serialization
    - tag may not match current version in config
    """

    def test_uses_tag_when_missing_current_version(self, scm_config: SCMConfig):
        """When the config does not have a current_version, the last tag is used."""
        # Arrange

        scm_info = SCMInfo(scm_config)
        scm_info.current_version = "1.2.3"
        config = Mock(spec=Config, current_version=None, scm_info=scm_info)

        # Act
        result = check_current_version(config)

        # Assert
        assert result == "1.2.3"



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from click import UsageError

from bumpversion import exceptions
from bumpversion.context import get_context
from bumpversion.versioning.version_config import VersionConfig
from tests.conftest import get_config_data


def test_invalid_parse_raises_error():
    """An invalid parse regex value raises an error."""
    with pytest.raises(exceptions.UsageError):
        VersionConfig(r"(.+", ("{major}.{minor}.{patch}",), search="", replace="")


class TestParse:
    """Tests for parsing a version."""

    def test_cant_parse_returns_none(self):
        """The default behavior when unable to parse a string is to return None."""
        # Arrange
        overrides = {
            "current_version": "19.6.0",
            "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
            "serialize": ["{major}.{minor}.{patch}"],
        }
        _, version_config, current_version = get_config_data(overrides)

        # Act and Assert
        assert version_config.parse("A.B.C") is None

    def test_cant_parse_raises_error_when_set(self):
        """When `raise_error` is enabled, the inability to parse a string will raise an error."""
        # Arrange
        overrides = {
            "current_version": "19.6.0",
            "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
            "serialize": ["{major}.{minor}.{patch}"],
        }
        _, version_config, current_version = get_config_data(overrides)

        # Act and Assert
        with pytest.raises(exceptions.UsageError):
            version_config.parse("A.B.C", raise_error=True)


class TestSerialize:
    """Tests for the VersionConfig.serialize() method."""

    def test_distance_to_latest_tag_in_pattern(self):
        """Using ``distance_to_latest_tag`` in the serialization string outputs correctly."""
        from bumpversion.scm.models import SCMConfig, SCMInfo

        overrides = {
            "current_version": "19.6.0",
            "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
            "serialize": ["{major}.{minor}.{patch}-pre{distance_to_latest_tag}"],
        }
        conf, version_config, current_version = get_config_data(overrides)
        scm_config = SCMConfig.from_config(conf)
        scm_info = SCMInfo(scm_config)
        scm_info.current_version = "19.6.0"
        scm_info.distance_to_latest_tag = 3
        scm_info.commit_sha = "1234123412341234"
        conf.scm_info = scm_info
        assert version_config.serialize(current_version, get_context(conf)) == "19.6.0-pre3"

    def test_environment_var_in_serialize_pattern(self):
        """Environment variables are serialized correctly."""
        import os

        os.environ["BUILD_NUMBER"] = "567"
        overrides = {
            "current_version": "2.3.4",
            "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
            "serialize": ["{major}.{minor}.{patch}-pre{$BUILD_NUMBER}"],
        }
        conf, version_config, current_version = get_config_data(overrides)
        assert version_config.serialize(current_version, get_context(conf)) == "2.3.4-pre567"
        del os.environ["BUILD_NUMBER"]



"""Tests for bumpversion.context."""

import os
from unittest.mock import patch

import pytest
from pytest import param

from bumpversion.context import calver_string_to_regex, prefixed_environ


@pytest.mark.parametrize(
    ["calver_format", "expected_regex"],
    [
        param("{YYYY}.{MM}", r"(?:[1-9][0-9]{3}).(?:1[0-2]|[1-9])", id="year-month"),
        param(
            "{YYYY}.{MM}.{DD}", r"(?:[1-9][0-9]{3}).(?:1[0-2]|[1-9]).(?:3[0-1]|[1-2][0-9]|[1-9])", id="year-month-day"
        ),
        param("{YYYY}.{WW}", r"(?:[1-9][0-9]{3}).(?:5[0-3]|[1-4][0-9]|[0-9])", id="year-week"),
        param("", "", id="empty"),
        param("version-2.0", "version-2.0", id="no-placeholders"),
        param(
            "v{YYYY}-release-{MM}", r"v(?:[1-9][0-9]{3})-release-(?:1[0-2]|[1-9])", id="mixed-text-and-placeholders"
        ),
        param("{invalid}", "{invalid}", id="invalid-placeholder"),
    ],
)
def test_calver_string_to_regex_replacements(calver_format: str, expected_regex: str) -> None:
    """The function should return the expected regex for the given CalVer format string."""
    # Act
    result = calver_string_to_regex(calver_format)
    # Assert
    assert result == expected_regex


class TestPrefixedEnviron:
    """Tests for the prefixed_environ function."""

    def test_prefixed_environ_with_mocked_environment(self):
        """Test that prefixed_environ returns a dictionary with prefixed keys."""
        # Arrange
        mocked_environ = {"KEY1": "value1", "KEY2": "value2"}
        expected_result = {"$KEY1": "value1", "$KEY2": "value2"}

        with patch.dict(os.environ, mocked_environ, clear=True):
            # Act
            result = prefixed_environ()

            # Assert
            assert result == expected_result

    def test_prefixed_environ_with_empty_environment(self):
        """Test that prefixed_environ works with an empty environment."""
        # Arrange
        mocked_environ = {}

        with patch.dict(os.environ, mocked_environ, clear=True):
            # Act
            result = prefixed_environ()

            # Assert
            assert result == {}

    def test_prefixed_environ_with_special_characters_in_keys(self):
        """Test that prefixed_environ handles special characters in keys."""
        # Arrange
        mocked_environ = {"KEY_WITH_$": "value1", "ANOTHER-KEY": "value2"}
        expected_result = {"$KEY_WITH_$": "value1", "$ANOTHER-KEY": "value2"}

        with patch.dict(os.environ, mocked_environ, clear=True):
            # Act
            result = prefixed_environ()

            # Assert
            assert result == expected_result



import subprocess
import sys

import pytest

from bumpversion.hooks import run_command


class TestRunCommand:
    """Test the run_command function."""

    def test_runs_a_str_command(self):
        """Runs the command formatted as a string."""
        result = run_command("echo Hello")
        assert isinstance(result, subprocess.CompletedProcess)
        assert result.stdout == "Hello\n"

    def test_can_access_env(self):
        """The command can access custom environment variables."""
        cmd = "echo %TEST_ENV%" if sys.platform == "win32" else "echo $TEST_ENV"
        result = run_command(cmd, environment={"TEST_ENV": "Hello"})
        assert isinstance(result, subprocess.CompletedProcess)
        assert result.stdout == "Hello\n"

    def test_non_zero_exit(self):
        """The result shows a non-zero result code."""
        result = run_command("exit 1")
        assert result.returncode == 1

    @pytest.mark.parametrize(
        "invalid_script",
        [(123,), (None,), (["exit", "1"])],
    )
    def test_an_invalid_script_raises_type_error(self, invalid_script):
        with pytest.raises(TypeError):
            run_command(invalid_script)

    @pytest.mark.parametrize(
        "invalid_env",
        [("string",), (123,), (None,)],
    )
    def test_an_invalid_env_raises_type_error(self, invalid_env):
        with pytest.raises(TypeError):
            run_command("echo Hello", environment=invalid_env)
